/*
Copyright (c) 2015  V.Ayllón.  All rights reserved.
 
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
public class Raster {
  //
  // Attention: Next  parameters must be the same than in printer Firmware
  //
  private static final float MAX_PCB_WIDTH = 80.5;    //Max width in mm including margin
  public static final int DPI = 600;                 // Dots per inch
  private static final float DPMM = DPI*(1.0/25.4);     // Dots per mm
  private static final int MAX_RASTER_BUFFER = 64;       // IMPORTANT: Must be power of two and 64 or less
  private static final int MAX_RAW_RASTER_BUFFER = 90;          // Max buffer we can process. Depends on MAX_RASTER_BUFFER
  //
  // These buffer and variables store the last line processed
  //
  private byte[] buffPixels =  new byte[MAX_RAW_RASTER_BUFFER+1];
  private int buffPosInit = 0;
  private int buffPosEnd = 0;
  private String outputString;
  private int rasterY = 0;
  private int rasterX = 0;
  private PImage psImage = null;
  private Base64 b64 = new Base64();


  private void setBitTrue(int bit) {      /* Set a specific bit to be True */
    if (bit/8 < MAX_RASTER_BUFFER)
      buffPixels[bit / 8] |= (1 << (bit % 8));
  }

  private void setBitFalse(int bit) {      /* Set a specific bit to be False */
    if (bit/8 < MAX_RASTER_BUFFER)
      buffPixels[bit / 8] &= ~(1 << (bit % 8));
  }

  //////////// Constructor

  Raster() {
  }
  /** 
   ** updateGraphics() - Updates the image
   **/
  public void updateGraphics(PGraphics pg)
  {
    if (psImage != null)
      pg.image(psImage, 0-psImage.width/2, 0-((psImage.height/2)+150));
  }
  /**
   ** load() - Load the filename
   **/
  public boolean load(String imgFileName) {
    rasterY = 0;
    rasterX = 0;
    psImage = null;
    psImage = cropImage(loadImage(imgFileName));
    //  psImage = loadImage(PSTEMPFILENAME);
    psImage.loadPixels();
    if (psImage.width/DPMM > MAX_PCB_WIDTH) {
      printlnMsg("Error: PCB exceed maximum width: " + psImage.width/DPMM  + ":"  + MAX_PCB_WIDTH + " mm");
      return true;
    }
    printlnMsg("File loaded: "+  String.format("%2.1f", psImage.width/DPMM) +"mm X " + String.format("%2.1f", psImage.height/DPMM) + "mm");
    // if invert flag checked invert image
    if (invCtrl.getItem(0).getValue() > 0)
      invertImage();
    return false;
  }
  /**
   ** isLoaded() - Return true is a file was loaded
   **/
  public boolean isLoaded() {

    if (psImage == null)
      return false;
    else 
      return true;
  }
  /**
   ** getWidth() - Return the image width
   **/
  public int getWidth() {

    if (psImage != null)
      return psImage.width;
    else 
      return 0;
  }
  /**
   ** getHeight() - Return the image height
   **/
  public int getHeight() {

    if (psImage != null)
      return psImage.height;
    else 
      return 0;
  }
  /**
   ** initRaster() - init the rastering
   **/
  public boolean initRastering(String imgFileName) {
    int d;
    if (load(imgFileName))
      return true;
    d = int(laserOnCtrl.getItem(0).getValue());
    if (d > 0)
      laserOnCtrl.getItem(0).setValue(0);
    d = int(laserDiodeCtrl.getValue());
    if (d <= 0) d = BLUE_DIODE;  
    initLaserMode(LASER_MODE_RASTER, d, getWidth(), getHeight());
    return false;
  }
  /**
   ** reset() - reset the class
   **/
  public void reset() {
    rasterY = 0;
    rasterX = 0;
    psImage = null;
    buffPosInit = 0;
    buffPosEnd = 0;

    // Turn laser off
    int d = int(laserOnCtrl.getItem(0).getValue());
    if (d > 0)
      laserOnCtrl.getItem(0).setValue(0);
    // Select default laser
    SetLaserDiode(int(laserDiodeCtrl.getValue()));
  }
  /**
   ** imageToGcode() - Raster image and generates G7 command
   **/
  public String imageToGcode() {
    int inx = 0;
    int loc = 0;
    int bitlen = 0;

    if (rasterY < psImage.height) {
      if (rasterY % 2 == 0) {  // Even lines
        for (inx = 0; inx < MAX_RASTER_BUFFER*8 && rasterX < psImage.width; inx++, rasterX++) {
          loc = rasterX + rasterY*psImage.width;
          if (inx == 0) buffPosInit = loc;
          if ((byte)psImage.pixels[loc] != 0) { 
            setBitFalse(inx);
          } else {
            setBitTrue(inx);
          }
        }
        buffPosEnd = loc;
        if (rasterX == psImage.width) {
          rasterX = psImage.width-1;
          rasterY++;
        }
        bitlen = max(1, ceil((float)inx/8.0));
        outputString = b64.encode(buffPixels, bitlen);
        return genG7Gcode(outputString, defLaserFeedrate, defLaserPower, b64.encodeLen(bitlen), 1);
      } else { // Odd lines
        for (inx = 0; inx < MAX_RASTER_BUFFER*8 && rasterX >=0; inx++, rasterX--) {
          loc = rasterX + rasterY*psImage.width;
          if (inx == 0) buffPosInit = loc;
          if ((byte)psImage.pixels[loc] != 0) { 
            setBitFalse(inx);
          } else {
            setBitTrue(inx);
          }
        }
        buffPosEnd = loc;
        if (rasterX < 0) {
          rasterX = 0;
          rasterY++;
        }
        bitlen = max(1, ceil((float)inx/8.0));
        outputString = b64.encode(buffPixels, bitlen);
        return genG7Gcode(outputString, defLaserFeedrate, defLaserPower, b64.encodeLen(bitlen), 0);
      }
    } else return null;
  }

  /**
   ** drawToCurrentLine() - Raster image and change the color till current line
   **/
  public void drawToCurrentLine(color c) {

    color rastColor = BLACK;

    if (buffPosEnd < buffPosInit) {  //N1 lines
      for (int n = buffPosInit; n >= buffPosEnd; n--) {
        if (psImage.pixels[n] == rastColor)
          psImage.pixels[n] = c;
      }
    } else { // N0 Lines
      for (int n = buffPosInit; n <= buffPosEnd; n++) {
        if (psImage.pixels[n] == rastColor)
          psImage.pixels[n] = c;
      }
    }
    psImage.updatePixels();
  }

  /**
   ** invertImage() - Inverts the image
   **/
  public void invertImage() {
    if (psImage != null) {
      psImage.filter(INVERT);
      sim.refresh();
    }
  }

  /**
   ** cropImage() - Crop a image to its limits
   **/
  private PImage cropImage(PImage img) {

    int loc;
    int topLimit = 0;
    int botLimit = 0;
    int leftLimit = 0;
    int rightLimit = 0;

    boolean found;

    // Find top limit
    found = false;
    for (int y = 0; y < img.height && found == false; y++) {
      for (int x = 0; x < img.width; x++) {
        loc = x + y*img.width;
        if (img.pixels[loc] != color (255, 255, 255)) {
          topLimit = (y > 0)? y-DEF_MATERIAL_MARGIN: y;
          found = true;
          break;
        }
      }
    }
    // Find bottom limit
    if (found) {
      found = false;
      // Find bottom limit
      for (int y = img.height-1; y >= 0 && found == false; y--) {
        for (int x = 0; x < img.width; x++) {
          loc = x + y*img.width;
          if (img.pixels[loc] != color (255, 255, 255)) {
            botLimit = (y < height-2)? y+DEF_MATERIAL_MARGIN: y;
            found = true;
            break;
          }
        }
      }
    } else { // Empty image
      return img;
    }
    // Find left limit
    found = false;
    for (int x = 0; x < img.width && found == false; x++) {
      for (int y = 0; y < img.height; y++) {
        loc = x + y*img.width;
        if (img.pixels[loc] != color (255, 255, 255)) {
          leftLimit = (x > 0)? x-DEF_MATERIAL_MARGIN: x;
          found = true;
          break;
        }
      }
    }
    // Find right limit
    if (found) {
      found = false;
      // Find bottom limit
      for (int x = img.width-1; x >= 0 && found == false; x--) {
        for (int y = 0; y < img.height; y++) {
          loc = x + y*img.width;
          if (img.pixels[loc] != color (255, 255, 255)) {
            rightLimit = (x < width-2)? x+DEF_MATERIAL_MARGIN: x;
            found = true;
            break;
          }
        }
      }
    } else { // Empty image
      return img;
    }
    // Create new image
    return img.get(leftLimit, topLimit, (rightLimit-leftLimit)+1, (botLimit-topLimit)+1);
  }

  /**
   ** printHex()
   **/

  void printHex() {

    int x = 0;
    int inx = 0;
    int loc;
    int h = psImage.height;
    int w =  psImage.width;

    for (int y = 0; y < h; y++) {

      x = 0;
      do {
        for (inx = 0; inx < h && x < w; inx++, x++) {
          loc = x + y*w;
          print(hex((byte)psImage.pixels[loc])+" ");
        }
        println();
      } 
      while (x < w);
    }
  }
  //////// Private class Base64 /////////////

  private class Base64 {
    /*
     * Copyright (c) 2015 Adam Rudd.
     * See LICENSE for more information
     */

    private final byte[] b64_alphabet = {
      'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 
      'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
    };

    //// Constructor ////

    Base64() {
    }

    //// Methods  ////

    public String encode(byte[] input, int inputLen) {
      byte[] output = new byte[MAX_RAW_RASTER_BUFFER+1];
      int i = 0, j = 0;
      int encLen = 0;
      byte[] a3 = new byte[3];
      byte[] a4;
      int inp = 0;

      while (inputLen-- > 0) {
        a3[i++] = input[inp++];
        if (i == 3) {
          a4 = a3_to_a4(a3);

          for (i = 0; i < 4; i++) {
            output[encLen++] = (byte)b64_alphabet[a4[i]];
          }
          i = 0;
        }
      }

      if (i > 0) {
        for (j = i; j < 3; j++) {
          a3[j] = '\0';
        }

        a4 = a3_to_a4(a3);

        for (j = 0; j < i + 1; j++) {
          output[encLen++] = (byte)b64_alphabet[a4[j]];
        }

        while ( (i++ < 3)) {
          output[encLen++] = '=';
        }
      }
      output[encLen] = '\0';
      return new String(output);
    }


    public int encodeLen(int plainLen) {
      int n = plainLen;
      return (n + 2 - ((n + 2) % 3)) / 3 * 4;
    }

    public int decodeLen(byte[] input, int inputLen) {
      int i = 0;
      int numEq = 0;
      for (i = inputLen - 1; input[i] == '='; i--) {
        numEq++;
      }
      return ((6 * inputLen) / 8) - numEq;
    }

    private byte[] a3_to_a4(byte[] a3) {
      byte[] ac = new byte[4];
      ac[0] = (byte)((a3[0] & 0xfc) >> 2);
      ac[1] = (byte)(((a3[0] & 0x03) << 4) + ((a3[1] & 0xf0) >> 4));
      ac[2] = (byte)(((a3[1] & 0x0f) << 2) + ((a3[2] & 0xc0) >> 6));
      ac[3] = (byte)((a3[2] & 0x3f));
      return ac;
    }
  }
}